/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/grpc_client/proto_wrap.hh"
#include "src/exception.hh"

#include "cdnskey_processor_api/common_types.pb.h"

namespace Fred {
namespace Akm {

namespace Api = CdnskeyProcessor::Api;

namespace {

template <typename> struct TypeTraits;

template <typename SrcType>
void proto_wrap(const SrcType& src, typename TypeTraits<SrcType>::ProtoType* dst);

template <>
struct TypeTraits<DomainFqdn>
{
    using ProtoType = Api::DomainFqdn;
};

template <>
struct TypeTraits<ScanTasks::value_type>
{
    using ProtoType = Api::Supervisor::ScanTask;
};

template <>
struct TypeTraits<DomainUuid>
{
    using ProtoType = Api::DomainUid;
};

template<>
void proto_wrap(const DomainFqdn& src, Api::DomainFqdn* dst)
{
    dst->set_value(*src);
}

Api::ScanType proto_wrap(ScanType src)
{
    switch (src)
    {
        case ScanType::insecure:
            return Api::ScanType::INSECURE;
        case ScanType::secure:
            return Api::ScanType::SECURE;
        case ScanType::secure_auto:
            return Api::ScanType::SECURE;
    }
    throw WrapperError{"unexpected ScanType"};
}

template<>
void proto_wrap(const DomainUuid& _src, Api::DomainUid* _dst)
{
    _dst->set_value(to_string(_src));
}

void proto_wrap(const DomainFqdn& fqdn, const ScanTaskInfo& info, Api::Domain* dst)
{
    proto_wrap(fqdn, dst->mutable_fqdn());
    proto_wrap(info.uuid, dst->mutable_uid());
    dst->set_scan_type(proto_wrap(info.scan_type));
}

template<>
void proto_wrap(const ScanTasks::value_type& src, Api::Supervisor::ScanTask* dst)
{
    proto_wrap(src.first, src.second, dst->mutable_domain());
    dst->mutable_nameservers()->Reserve(src.second.nameservers.size());
    for (const auto& nameserver : src.second.nameservers)
    {
        dst->add_nameservers()->set_value(*nameserver);
    }
}

} // namespace Fred::Akm::{anonymous}

CdnskeyProcessor::Api::Supervisor::ScanTaskCollectionRequest proto_wrap(const ScanTasks::const_iterator& begin, const ScanTasks::const_iterator& end)
{
    CdnskeyProcessor::Api::Supervisor::ScanTaskCollectionRequest dst;
    for (auto task_iter = begin; task_iter != end; ++task_iter)
    {
        proto_wrap(*task_iter, dst.add_items());
    }
    return dst;
}

} // namespace Fred::Akm
} // namespace Fred
